import sqlite3
import time
from server.lib import logger
from server.lib import config


STATUS = {"PENDING", "QUEUED",
          "EXECUTING", "COMPLETED",
          "ERROR", "ABORTED",
          "UNKNOWN", "HELD",
          "SUSPENDED", "ARCHIVED"}

JOBLIST_TYPE = {"PLOT": "PLOT",
                "EXTRACT": "EXTRACT"
                }


class JobsInterface(object):

    def __init__(self):
        self.dbName = config.JOBS_DB
        self.connection = None
        self.cursor = None
        self.connected = False

    def connect(self):
        if self.connected is False:
            self.connection = sqlite3.connect(self.dbName)
            self.connection.row_factory = sqlite3.Row
            self.cursor = self.connection.cursor()
            self.connected = True

    def disconnect(self):
        if self.connected is True:
            self.connection.close()
            self.connected = False

    def add_new_joblist(self, list_type):
        if list_type in JOBLIST_TYPE:
            request = "INSERT INTO t_joblist (creation_time, type) VALUES (?, ?)"
            self.cursor.execute(request, (str(time.time()), list_type, ))
            self.connection.commit()
            return self.cursor.lastrowid
        else:
            raise Exception("Unknown joblist type %s" % list_type)

    def add_new_job(self, id_joblist, owner, parameters):
        request = "INSERT  INTO t_job (id_joblist, owner, parameters, phase) \
                   VALUES (?, ?, ?, ?)"
        self.cursor.execute(request,
                            (id_joblist, owner, parameters, 'PENDING'))
        self.connection.commit()
        return self.cursor.lastrowid

    def change_job_status(self, id_job, new_status):
        if new_status in STATUS:
            request = "UPDATE t_job \
                       SET phase = ? \
                       WHERE id = ?"
            self.cursor.execute(request, (new_status, id_job))
            self.connection.commit()
        else:
            raise Exception("Try to set unknown job status")

    def set_error_status(self, id_job, error):
        request = "UPDATE t_job \
                   SET phase = 'ERROR', error = ? \
                   WHERE id = ?"
        self.cursor.execute(request, (error, id_job))
        self.connection.commit()

    def set_completed_status(self, id_job, result):
        request = "UPDATE t_job \
                   SET phase = 'COMPLETED', result = ? \
                   WHERE id = ?"
        self.cursor.execute(request, (result, id_job))
        self.connection.commit()

    def get_next_job(self):
        request = "SELECT j.id, j.parameters, j.owner, jl.type \
                   FROM t_job AS j \
                   JOIN t_joblist AS jl ON j.id_joblist = jl.id \
                   WHERE j.phase = 'PENDING' \
                   ORDER BY j.id ASC \
                   LIMIT 1"
        self.cursor.execute(request)
        return self.cursor.fetchone()

    def get_jobs_in_joblist(self, id_joblist):
        request = "SELECT id, parameters, owner, phase, result \
                   FROM t_job \
                   WHERE id_joblist = ? \
                   ORDER BY id ASC "
        self.cursor.execute(request, (id_joblist,))
        return self.cursor.fetchall()

    def get_job_by_id(self, id_job):
        request = "SELECT j.id, j.parameters, j.owner, \
                   j.phase, j.result, j.error, jl.type \
                   FROM t_job j \
                   JOIN t_joblist AS jl ON j.id_joblist = jl.id \
                   WHERE j.id = ?"
        self.cursor.execute(request, (id_job,))
        return self.cursor.fetchone()

    def get_jobs_by_phase(self, phase):
        request = " SELECT * \
                    FROM t_job AS j \
                    JOIN t_joblist AS jl ON j.id_joblist = jl.id \
                    WHERE phase = ?"
        self.cursor.execute(request, (phase, ))
        return self.cursor.fetchall()

    def delete_job_by_id(self, id_job):
        request = "DELETE FROM t_job WHERE id = ?"

        self.cursor.execute(request, (id_job, ))
        self.connection.commit()

    def purge_jobs(self):
        # delete jobs
        request = " DELETE FROM t_job"
        self.cursor.execute(request)
        self.connection.commit()
        # reset id autoincrement count
        request = "DELETE FROM sqlite_sequence WHERE name = 't_job'"
        self.cursor.execute(request)
        self.connection.commit()


    def purge_joblists(self):
        # delete joblists
        request = " DELETE FROM t_joblist"
        self.cursor.execute(request)
        self.connection.commit()
        # reset id autoincrement count
        request = "DELETE FROM sqlite_sequence WHERE name = 't_joblist'"
        self.cursor.execute(request)
        self.connection.commit()

    def purge_sessions(self):
        """
            duration in seconds
        """
        # removing sessions
        request = " DELETE FROM t_session"
        self.cursor.execute(request)
        self.connection.commit()
        # reset id autoincrement count
        request = "DELETE FROM sqlite_sequence WHERE name = 't_session'"
        self.cursor.execute(request)
        self.connection.commit()

    def add_session(self, uuid, creation_date):
        request = "INSERT INTO t_session (uuid, creation_date) VALUES (?, ?)"
        self.cursor.execute(request, (uuid, creation_date, ))
        self.connection.commit()
        return self.cursor.lastrowid

    def get_session_by_uuid(self, uuid):
        request = "SELECT * FROM t_session WHERE uuid = ?"
        self.cursor.execute(request, (uuid, ))
        return self.cursor.fetchone()

    def set_session_data(self, uuid, data):
        request = "UPDATE t_session SET data = ? WHERE uuid = ?"
        self.cursor.execute(request, (data, uuid, ))
        self.connection.commit()
