# -*- coding: utf-8 -*-
import h5py as h5
import numpy as numpy
from server.lib.logger  import append_info, append_exception

__author__ = "Nicolas Moreau"
__copyright__ = "Copyright 2012"
__credits__ = ["Nicolas Moreau"]
__license__ = "GPL"
__version__ = "1.0"
__maintainer__ = "Nicolas Moreau"
__email__ = "nicolas.moreau@obspm.fr"
__status__ = "Development"


'''
Notes :

PS: Table Metadata :
 Metadata contains :
PathHDF5 | DataSet | Col.Index | PubID | HName  | DataType | Unit | SKOS | UCD | utype | Description | ObjParent | Group |
   0     |    1    |     2     |   3   |   4    |    5     |  6   |  7   |  8  |  9    |   10        |    11     |  12   |
   
 Metadata_ObjType contains :
 PubID | Hname | SKOS | Description

'''



class PointerFactory(object):
    """
    creates a Pointer object
    """
    def __init__(self):
        self.pointer = Pointer()

    def withPath(self, path):
        self.pointer.path = path
        return self

    def withDataset(self, dataset):
        self.pointer.dataset = dataset
        return self

    def withName(self, name):
        self.pointer.name = name
        return self

    def withColumn(self, column):
        self.pointer.column = column
        return self

    def withUnit(self, unit):
        self.pointer.unit = unit
        return self

    def withUcd(self, ucd):
        self.pointer.ucd = ucd
        return self

    def withSkos(self, skos):
        self.pointer.skos = skos
        return self

    def withUtype(self, utype):
        self.pointer.utype = utype
        return self

    def withDescription(self, description):
        description = description.strip()
        if description != '' and description != "none":
            self.pointer.description = description.strip()
        return self

    def withObjParent(self, parent):
        self.pointer.objParent = parent
        return self

    def withGroup(self, group):
        self.pointer.group = group
        return self

    def getPointer(self):
        return self.pointer

    def getErrorPointer(self, quantity):
        self.pointer = Pointer()
        self.pointer.name = quantity
        self.pointer.unit = "N/A"
        self.pointer.description = "error"
        return self.pointer


class Pointer(object):
    """
        Points on data in a dataset
    """
    def __init__(self):
        # dataset path
        self.path = None
        # dataset name
        self.dataset = None
        # quantity human readable name
        self.name = None
        # column position in dataset
        self.column = None
        self.ucd = None
        self.skos = None
        self.utype = None
        self.unit = None
        self.description = None
        self.objParent = None
        self.group = None


class PdrHDF(object):
    """
      Access to a pdr HDF5 file
    """
    def __init__(self, filename, treeManager=None):
        # hdf file
        self.f = h5.File(filename, 'r')
        """
        create a pivoted array with the metadata
         ------------------------------------------------------------------------------------
         The old versions of numpy do not treat automatically the fact that
         the metadata table is composed of string of variable lengths.
         Beside updating numpy, another solution consists in setting the type
         to "object" when creating the numpy array
         see http://stackoverflow.com/questions/16037824/how-to-convert-numpy-object-array-into-str-unicode-array
         note: setting the type to unicode or post-converting
         object type via astype('U') does not work either
         with the old version of numpy
         ------------------------------------------------------------------------------------
        """
        self.metadata = numpy.array(self.f['/Metadata/Metadata'], 
                                    dtype=object)   
        self.xQuantitiesPath = "/Local quantities/Positions"
        # separator in path
        self.separator = "#"
        # show only local quantites directory to user
        # ~ self.selectableData = "/Local quantities"

        self.treeManager = treeManager

        # indexes all the data
        init_result = self._buildIndex()       
        self.xQuantities = init_result["xquantities"]
        self.index = init_result["pointers"]
        # ~ if treeManager is not None :             
            # ~ self.tree = init_result["tree"]
        # ~ else : 
            # ~ self.tree = None

    def close(self):
        """
            close hdf file
        """
        self.f.close()

    def getByGroup(self, group):
        """
            returns the object located at group position
        """
        try:
            return self.f[group]
        except KeyError:
            print("group : " + group + " not found")

    def getDefaultColumns(self):
        result = []
        for key in self.index['Positions']:
            result.append(key)
        return result

    def _buildIndex(self):
        """
            creates the index
        """
        result = {}
        # map of Pointer objects
        pointers = {}  
        # quantities that can be used as x axis on result plot
        result["xquantities"] = [] 

        for line in self.metadata:
            line_utf = []
            # cast everything to string to avoid problems in gui
            # when comparing used input and index content
            # performance pust be checked on big files
            for value in line:
                line_utf.append(value.decode('UTF-8'))

            if line_utf[1] not in pointers:
                pointers[line_utf[1]] = {}

            pf = PointerFactory()
            pf.withPath(line_utf[0]).withDataset(line_utf[1])
            pf.withName(line_utf[4]).withColumn(line_utf[2])
            pf.withUnit(line_utf[6]).withUcd(line_utf[8])
            pf.withSkos(line_utf[7]).withUtype(line_utf[9])
            pf.withDescription(line_utf[10]).withObjParent(line_utf[11])
            pf.withGroup(line_utf[12]).getPointer()            
            
            # pointer is identified by the human readable name of a quantity
            index_name = line_utf[4]
            if isinstance(index_name, str) is False:
                raise Exception("%s is not a string " % index_name)
            pointers[index_name] = pf.getPointer()

            if self.treeManager is not None : 
                self.treeManager.addToTree(line_utf, pointers, self.separator)
            
            if line_utf[0] + '/' + line_utf[1] == self.xQuantitiesPath:
                result['xquantities'].append(line_utf[4])

            result["pointers"] = pointers
            # ~ if self.treeManager is not None : 
                # ~ result["tree"] = self.treeManager.tree[0]

        return result

    def extractQuantity(self, p):
        """
            extract a quantity according to the pointer p
        """
        ds = self.f["%s/%s" % (p.path, p.dataset)]
        # 1D array, returned directly
        if len(ds.shape) == 1:
            return ds
        else:
            return ds[:, int(p.column)]
    
    
    def getCode(self):
        """
            return version of the code (pdr) as string or empty
            string if property is not found
        """
        if "Code version" in self.index:
            return self.extractQuantity(self.index["Code version"])[0]\
                                            .decode("UTF-8")
        else:
            return ""


