#!flask/bin/python
import os.path
import uuid
import yaml
import pickle
import time
from flask import Flask, Response, jsonify, request, json, send_from_directory, make_response, redirect
from flask_caching import Cache
from werkzeug.utils import secure_filename
from server.lib import database as jobdb
from server.lib import logger as logger
from server.lib import hdf as hdf
from server.lib import hdfutil as hdfutil
from ismvoca.lib import ismvoca as ia
from server.lib.config import *
from server.lib import yamlbuilder as yaml

try:
    from cheroot.wsgi import Server as WSGIServer, PathInfoDispatcher
except ImportError:
    from cherrypy.wsgiserver import CherryPyWSGIServer as WSGIServer, WSGIPathInfoDispatcher as PathInfoDispatche


cache_config = {
    "DEBUG": False,
    "CACHE_TYPE": "simple",
    "CACHE_DEFAULT_TIMEOUT": 300
}

app = Flask(__name__)
app.config.from_mapping(cache_config)
cache = Cache(app)
app.secret_key = b'\x05\xf5\x9fsH\x0e\xb3\x9f\xb7\xa6\x05\x8dQ\x06\x18\xd8'

d = PathInfoDispatcher({'/':app})
server = WSGIServer(('0.0.0.0', SERVER_PORT), d)

concepts_db = None
ngdb = None

@cache.cached(timeout=360000000, key_prefix='autompletion_vocab')
def load_concepts_file(code):
    """
        load the concepts file and store result into local cache
    """
    concepts_db = None
    ngdb = None

    if "pdr" in code.lower():            
        qnbinary = os.path.join(QNAMES_FILE_DIRECTORY, "pdr" + QNAMES_BINARY_SUFFIX)
        qnfile =  os.path.join(QNAMES_FILE_DIRECTORY, "pdr" + QNAMES_FILE_SUFFIX)
    else:
        qnbinary = os.path.join(QNAMES_FILE_DIRECTORY, "shock" + QNAMES_BINARY_SUFFIX)
        qnfile =  os.path.join(QNAMES_FILE_DIRECTORY, "shock" + QNAMES_FILE_SUFFIX)

    try:
        if os.path.isfile(qnbinary):
            with open(qnbinary, 'rb') as dbf:
                concepts_db = pickle.load(dbf)
        else:
            concepts_db = ia.load_concepts_db(qnfile, yaml)
            with open(qnbinary, "wb") as dbf:
                pickle.dump(concepts_db, dbf)
        ngdb = ia.build_nghash(ia.cdb_hnames(concepts_db), hf=None)
        return concepts_db, ngdb

    except Exception as e:
        logger.append_exception(e)
        raise e

def get_session_data(s_uuid):
    db_int = jobdb.JobsInterface()
    db_int.connect()
    result = db_int.get_session_by_uuid(s_uuid)
    session_data = json.loads(result['data'])
    db_int.disconnect()
    return session_data


def bad_request(message):
    """Handles request error
    :param message: message explaining the error
    :type message: string

    """
    response = jsonify({'message': message})
    response.status_code = 400
    return response


@app.route(BASE_URL + "/")
def index():
    """Homepage of the service

    :return: main page
    :rtype: string
    """
    return "ISM service"
    
@app.route(BASE_URL + "/projectload/<project>/<model>", methods=["GET"])
def projectload(project, model):
    """Load a file from its name and origin project
       Store it in session a correspondance between the file and
       a session id
    
    Used in web mode

    :param project: name of the project containing the datafile
    :type project: string

    :param model: name of the file to open in project containing the datafile
    :type model: string    

    :return: a map describing the structure of the file and the session id
    :rtype: json
    """
    return load(model, project)

@app.route(BASE_URL + "/localload/<filename>", methods=["GET"])
def localload(filename):
    """Load a file only from its name ( in data directory)
       Stores it in session a correspondance between the file and
       a session id
       
       Used in local mode

    :param filename: name of the hdf file to open
    :type filename: string
    :return: a map describing the structure of the file and the session id
    :rtype: json
    """
    return load(filename)        
        
def load(model, project=None):
    global concepts_db, ngdb
    if project is None:
        path = os.path.join(DATA_DIRECTORY, model)
        filename = model
    else:
        path = os.path.join(DATA_DIRECTORY, project, "raw", model, model+"_s_20.hdf5")
        filename = os.path.join(project, "raw", model, model+"_s_20.hdf5")
    try:
        hdf_file = hdf.PdrHDF(path)
        tree = hdfutil.TreeManager(hdf_file)   
        session_id = str(uuid.uuid4())
        concepts_db, ngdb = load_concepts_file(hdf_file.getCode())
        db_int = jobdb.JobsInterface()
        db_int.connect()
        db_int.add_session(session_id, time.time())
        db_int.set_session_data(session_id, json.dumps({'filename': filename}))
        hdf_file.close()
        db_int.disconnect()
        response = jsonify({'tree': tree.tree,
                            "sessionId": session_id,
                            "xQuantities": hdf_file.xQuantities,
                            "xDefault": "AV"})
        return response
    except OSError as e:
        print(e)
        return bad_request("Unable to open {}".format(filename))   
    except Exception as e:
        print(e)
        return bad_request(e)   

    
        
def allowed_file(filename):    
    return '.' in filename and \
        filename.rsplit('.', 1)[1].lower() in ALLOWED_EXTENSIONS

@app.route(BASE_URL + '/import', methods=['POST'])
def upload_file():
    # check if the post request has the file part
    if 'file' not in request.files:        
        return "No file to upload"
    file = request.files['file']
    if file.filename == '':
        return 'No selected file'
    if file and allowed_file(file.filename):
        try:
            filename = secure_filename(file.filename)
            # if file is already in directory, do not upload it
            if os.path.isfile(os.path.join(DATA_DIRECTORY, filename)) is False:
                file.save(os.path.join(DATA_DIRECTORY, filename))
            return redirect(f"http://localhost:{SERVER_PORT}{BASE_URL}/html/index.html?model={filename}",
                            code=302)
        except Exception as e:
            print(e)
    return "file type is not allowed"


@app.route(BASE_URL + "/export/data/text", methods=["POST"])
def export_as_text():
    """Export data in text format

    :param json: data to export
    :type json: json
    """
    session_id = request.json["sessionId"]
    session_data = get_session_data(session_id)
    
    parameters = {"parameters" : request.json["quantities"],
                  "datatype": request.json["datatype"],
                  "files": (session_data['filename'],)}
    
    job_id = write_configuration("text", parameters)
    response = jsonify({"job_id": job_id, "status": "OK"})
    return response


@app.route(BASE_URL + "/export/data/votable", methods=["POST"])
def export_as_votable():
    """Export data in votable format

    :param json: data to export
    :type json: json
    """
    session_id = request.json["sessionId"]
    session_data = get_session_data(session_id)
    
    parameters = {"parameters" : request.json["quantities"],
                  "datatype": request.json["datatype"],
                  "files": (session_data['filename'],)}
    job_id = write_configuration("votable", parameters)
    response = jsonify({"job_id": job_id, "status": "OK"})
    return response


@app.route(BASE_URL + "/export/quantities", methods=['POST'])
def export_quantities():
    """Export a list of quantities in text format and return it in a file

    :param node_selection: quantities
    :type node_selection: list
    """
    quantities = request.json["quantities"]
    if len(quantities) > 0:
        filename = str(uuid.uuid4()) + '.txt'
        with open(OUTPUT_DIRECTORY + '/' + filename, 'w') as f:
            f.write("\n".join(quantities))
        response = jsonify({"filename": filename})
        return response
    else:
        bad_request("There was no quantity in the request.")


@app.route(BASE_URL + "/execute/extract", methods=["POST"])
def execute_export_script():
    """Apply a quantity extraction to a list of files

    :param node_selection: quantities
    :type node_selection: list
    """
    json_content = request.get_json()    
    job_id = write_configuration("text", json_content)        
    response = jsonify({"job_id": job_id, "status": "OK"})
    return response
    
@app.route(BASE_URL + "/execute/plot", methods=["POST"])
def execute_plot_script():
    """Apply a quantity extraction to a list of files

    :param node_selection: quantities
    :type node_selection: list
    """
    try : 
        json_content = request.get_json()    
        job_id = write_plot_configuration(json_content, json_content['files'])        
        response = jsonify({"job_id": job_id, "status": "OK"})
        return response
    except Exception as e:
        print(e)


@app.route(BASE_URL + "/autocomplete/<usersearch>", methods=["GET"])
def autocomplete(usersearch):
    """Endpoint providing autocompletion
    :param usersearch: a text input for which we will provide suggestions
    :type usersearch: string
    """
    concepts_db, ngdb = load_concepts_file()
    try:
        cfilters = (
            ia.cf_contains, 
            ia.mk_cf_likelyhood_ss_sort(group_filter_compress_size = 4),
            ia.mk_cf_sample_len(4),
            ia.mk_cf_sample_compress(SUGGEST_RESULT_LIMIT)
        )

        ac = ia.autocomplete(concepts_db, ngdb, usersearch, cfilters)
    except Exception as e:
        print(e)
        ac = ["not found"]

    response = jsonify(ac)
    return response


@app.route(BASE_URL + "/plot", methods=["POST"])
def plot():
    """Plot data

    :param json: data to export
    :type json: json
    """
    try:
        plot_config = request.json["parameters"]
        session_id = request.json["sessionId"]
        session_data = get_session_data(session_id)
        if len(plot_config['y1']) > 0:
            job_id = write_plot_configuration(plot_config,
                                              (session_data['filename'],))
            response = jsonify({"job_id": job_id, "status": "OK"})
            return response
        else:
            response = jsonify({"message": "Y1 list is empty",
                                "status": "ERROR"})

    except Exception as e:
        logger.append_info(e)
        response = jsonify({"message": str(e), "status": "ERROR"})


def write_configuration(exportformat, parameters):
    quantities = parameters["parameters"]
    datatype = parameters["datatype"]
    hdffiles = parameters["files"]
    if "output_dir" in parameters :
        output_dir = parameters["output_dir"]
    else : 
        output_dir = None
    # id of created jobs
    result = []
    # creates a job list
    db_int = jobdb.JobsInterface()
    db_int.connect()
    joblist_id = db_int.add_new_joblist('EXTRACT')
    # add jobs into joblist
    for f in hdffiles:
        try :
            config = {
                'dataformat': exportformat,
                'datatype': datatype,
                'quantities': quantities,
                'output_path': output_dir,
                'hdffile': f
            }
            job_id = db_int.add_new_job(joblist_id,
                                        "test@obspm.fr",
                                        json.dumps(config))
            result.append(job_id)
        except Exception as e :
            print(e)
    db_int.disconnect()
    return result


def write_plot_configuration(params, hdffiles):
    def add_limit(conf_dict, limit, params):
        try:
            if params[limit].strip() != '':
                conf_dict[limit] = params[limit]
        except Exception as e:
            logger.append_exception(e)
    
    # id of created jobs
    result = []
    # creates a job list
    db_int = jobdb.JobsInterface()
    db_int.connect()
    joblist_id = db_int.add_new_joblist('PLOT')

    
    # yaml configuration has already been created (cli request)
    if 'yaml' in params:
        for f in hdffiles:
            d_uuid = str(uuid.uuid4())
            output = d_uuid + '_' +   f.replace('.hdf5', "").replace('/', '_')
            config = {
                'hdffile': f,
                'plotconfiguration': {
                    "yaml": params['yaml']
                },
                'output': output
            }     
            job_id = db_int.add_new_job(joblist_id,
                                        "test@obspm.fr",
                                        json.dumps(config))
            result.append(job_id)
    # yaml conf does not exist yet (web request)
    else:
        # add jobs into joblist
        for f in hdffiles:
            d_uuid = str(uuid.uuid4())
            # ~ output = d_uuid + '_' +   f.replace('.hdf5', "").replace('/', '_')
            output = f.replace('.hdf5', "").replace('/', '_')

            if 'y2' in params:
                y2 = params['y2']
                for val in y2:
                    val.append('--')
                    

            plotconfiguration = {
                "x": params['x'],
                "y1": params['y1'],
                "y2": y2,
                "xlabel": params['xlabel'],
                "y1label": params['y1label'],
                "y2label": params['y2label'],
                "log": params['log'],
                "title": params['title'],
                "output": output
            }

            add_limit(plotconfiguration, 'xmin', params)
            add_limit(plotconfiguration, 'xmax', params)
            add_limit(plotconfiguration, 'y1min', params)
            add_limit(plotconfiguration, 'y1max', params)
            add_limit(plotconfiguration, 'y2min', params)
            add_limit(plotconfiguration, 'y2max', params)

            config = {
                'hdffile': f,
                'plotconfiguration': {
                    'yaml' : yaml.generate_yaml(plotconfiguration)
                },
                'output': output                
            }

            job_id = db_int.add_new_job(joblist_id,
                                        "test@obspm.fr",
                                        json.dumps(config))
            result.append(job_id)
        db_int.disconnect()
        return result


"""
    pseudo-UWS API
"""


@app.route(BASE_URL + "/jobs/<job_id>", methods=["GET"])
def get_job(job_id):
    db_int = jobdb.JobsInterface()
    db_int.connect()
    job = db_int.get_job_by_id(int(job_id))
    try:
        if job is not None:
            result = {"id": job["id"],
                      "parameters": job['parameters'],
                      "owner": job["owner"],
                      "phase": job["phase"],
                      "result": job['result'],
                      "yaml": YAML_TEMPLATE_NAME,
                      "error": job["error"]}
            response = jsonify(result)
            return response
        else:
            response = jsonify([])
            return response
    except Exception as e:
        #! shoud return error to client side
        print(e)


"""
    !!! functions to use only in local environment
    !!! to remove when hosting on a web server
"""

@app.route(BASE_URL + '/output/<filename>')
def serve_static(filename):
    # filename starts by uuid
    parts = filename.split("_")
    # remove uuid from proposed file name
    parts.pop(0)    
    response = make_response(send_from_directory((OUTPUT_DIRECTORY), filename))
    response.headers['Content-Disposition'] = f'attachment; filename="{ "_".join(parts)}"'
    response.headers['Cache-Control'] = 'no-cache'
    return response
    

@app.route(BASE_URL + '/html/<filename>')
def serve_html(filename):
    if filename in {"index.html", 
                    "import.html", 
                    "help.html"}:
        return send_from_directory((CLIENT_DIRECTORY), filename)
    else:
        return "Unknown page"


@app.route(BASE_URL + '/html/assets/css/<filename>')
def css(filename):
    return send_from_directory((CLIENT_DIRECTORY + "/assets/css"), filename)


@app.route(BASE_URL + '/html/assets/js/<filename>')
def js(filename):
    return send_from_directory((CLIENT_DIRECTORY + "/assets/js"), filename)


@app.route(BASE_URL + '/html/assets/img/<filename>')
def img(filename):
    return send_from_directory((CLIENT_DIRECTORY + "/assets/img"), filename)


def listen():
    global server
    # ~ load_concepts_file()
    try:
        server.start()
    except KeyboardInterrupt:
        server.stop()



if __name__ == '__main__':
    listen()

